"""NAT plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_findone

RULE_NAMES = (
    "descr,disabled,interface,protocol,ipprotocol,source,"
    "sourceport,source_hash_key,destination,dstport,target,"
    "targetip,targetip_subnet,local-port,associated-rule-id,poolopts,"
    "updated,created"
)

NODE_NAMES = "direction,mode," + RULE_NAMES
WIDTHS = "20,20,20,60,20,20,40,20,30,30,20,40,20,20,20,60,30,60,60"


class Plugin(BasePlugin):
    """Gather data for the System Groups."""

    def __init__(
        self,
        display_name: str = "NAT",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)
        self.local_data = {}

    def adjust_node(self, node: Node) -> str:
        """Local node adjustments."""
        if node is None:
            return ""

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather information."""
        rows = []

        nat_node = xml_findone(parsed_xml, "nat")
        if not len(nat_node):
            return

        rule_names = RULE_NAMES.split(",")

        # No mode in inbound.
        self.local_data = dict(direction="inbound", mode="")

        # Cycle through top-level rules, which are inbound.
        rule_nodes = xml_findall(nat_node, "rule")

        for node in rule_nodes:
            self.report_unknown_node_elements(node, rule_names)
            row = []

            for node_name in self.node_names:
                if node_name in ("direction", "mode"):
                    row.append(self.local_data[node_name])
                    continue

                value = self.adjust_node(xml_findone(node, node_name))

                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        # Cycle through the outbound children's rules.

        outbound_node = xml_findone(nat_node, "outbound")
        self.local_data["direction"] = "outbound"
        self.local_data["mode"] = self.adjust_node(xml_findone(outbound_node, "mode"))
        row = []

        rule_nodes = xml_findall(outbound_node, "rule")
        for node in rule_nodes:
            self.report_unknown_node_elements(node, rule_names)
            row = []

            for node_name in self.node_names:
                if node_name in ("direction", "mode"):
                    row.append(self.local_data[node_name])
                    continue

                value = self.adjust_node(xml_findone(node, node_name))

                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        rows.sort()

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""HAProxy plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from textwrap import indent
from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData, split_commas
from ..support.elements import xml_findall, xml_findone


class Plugin(BasePlugin):
    """
    Gather HAProxy data.

    Generates multiple sheets of data:
      * overview
      * backends
      * pools
    """

    def __init__(
        self,
        display_name: str = "HAProxy",
        node_names: str = "",
        column_widths: str = "",
    ) -> None:
        """Ignore node_names and column_widths as we create them individually."""
        super().__init__(
            display_name,
            node_names,
            column_widths,
            ["pfsense,installedpackages,haproxy,advanced"],
        )

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Document haproxy configuration."""
        haproxy_node = xml_findone(parsed_xml, "installedpackages,haproxy")
        if haproxy_node is None:
            return

        for overview in self._overview(haproxy_node):
            yield overview

        ha_backends_nodes = xml_findall(haproxy_node, "ha_backends,item")
        for backend in self._backends(ha_backends_nodes):
            yield backend

        pools_nodes = xml_findall(haproxy_node, "ha_pools,item")
        for pool in self._pools(pools_nodes):
            yield pool

    def adjust_node(self, node: Node) -> str:
        """Local node adjustment."""
        if node is None:
            return ""

        # Custom nodes before standard
        match node.tag:
            case "a_extaddr":
                cells = []
                addr_nodes = xml_findall(node, "item")
                for a_node in addr_nodes:
                    cell = []
                    address = self.adjust_node(xml_findone(a_node, "extaddr"))
                    port = self.adjust_node(xml_findone(a_node, "extaddr_port"))
                    cell.append(f"{address}:{port}")
                    cell.append(
                        f"""ssl: {self.adjust_node(xml_findone(a_node, "extaddr_ssl"))}"""
                    )
                    cell.append(self.adjust_node(xml_findone(a_node, "_index")))
                    cells.append("\n".join(cell))

                return "\n".join(cells)

            case "dcertadv":
                # ssl-min-ver TLS ciphers x:x
                result = []
                value = node.text
                if not value:
                    return ""
                els = value.split(" ")
                if len(els) != 4:
                    raise NodeError(f"Unexpected tag value for {node.tag}: dcertadv")
                result.append(f"{els[0]}: {els[1]}")
                result.append("ciphers:")
                result.append(indent("\n".join(els[3].split(":")), " " * 4))
                return "\n".join(result)

            case "ha_servers":
                server_nodes = xml_findall(node, "item")
                rows = []
                for server_node in server_nodes:
                    row = []
                    row.append(
                        f"""name: {self.adjust_node(xml_findone(server_node, "name"))}"""
                    )
                    address = self.adjust_node(xml_findone(server_node, "address"))
                    port = self.adjust_node(xml_findone(server_node, "port"))
                    row.append(f"domain/port: {address}:{port}")
                    for node_name in "ssl,checkssl,id,_index".split(","):
                        value = self.adjust_node(xml_findone(server_node, node_name))
                        row.append(f"{node_name}: {value}")
                    row.append("")
                    rows.append("\n".join(row))

                return "\n".join(rows)

        return super().adjust_node(node)

    def _overview(self, node: Node) -> Generator[SheetData, None, None]:
        """
        Top-level haproxy elements.

        Display two columns (node, value).

        """
        node_names: list[str] = (
            "enable,configversion,nbproc,nbthread,maxconn,carpdev,"
            "logfacility,loglevel,log-send-hostname,remotesyslog,"
            "localstats_refreshtime,localstats_sticktable_refreshtime"
            ",dns_resolvers,resolver_retries,resolver_timeoutretry,resolver_holdvalid,"
            "hard_stop_after,ssldefaultdhparam,"
            "email_mailers,email_level,email_myhostname,email_from,email_to,"
            "config,files,advanced"
        ).split(",")
        header_row: list[str] = "name,value".split(",")
        column_widths: list[int] = split_commas("50,80")

        all_node_names = node_names[:]
        all_node_names.extend("ha_backends,ha_pools".split(","))
        self.report_unknown_node_elements(node, all_node_names)
        row = []

        for node_name in node_names:
            value = self.adjust_node(xml_findone(node, node_name))

            row.append(value)

        self.sanity_check_node_row(node, row)
        rows = list(zip(node_names, row))

        yield SheetData(
            sheet_name="HAProxy",
            header_row=header_row,
            data_rows=rows,
            column_widths=column_widths,
        )

    def _backends(self, nodes: list[Node]) -> Generator[SheetData, None, None]:
        """HAProxy backends have one or more items."""
        node_names: list[str] = split_commas(
            "name,status,type,primary_frontend,backend_serverpool,"  # 5
            "dontlognull,log-detailed,socket-stats,a_extaddr,ha_certificates,"  # 10
            "clientcert_ca,clientcert_crl,a_actionitems,a_errorfiles,dcertadv,"  # 15
            "ssloffloadcert,forwardfor,advanced,ha_acls,httpclose"  # 19
        )
        column_widths: list[int] = split_commas(
            "25,20,20,25,25,"  # 5
            "20,20,20,30,20,"  # 10
            "20,20,20,20,50,"  # 15
            "20,40,20,20,20"  # 19
        )

        rows = []

        for node in nodes:
            self.report_unknown_node_elements(node, node_names)
            row = []

            for node_name in node_names:
                value = self.adjust_node(xml_findone(node, node_name))

                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        yield SheetData(
            sheet_name="HAProxy Backends",
            header_row=node_names,
            data_rows=rows,
            column_widths=column_widths,
        )

    def _pools(self, nodes: list[Node]) -> Generator[SheetData, None, None]:
        """Report HAProxy pools."""
        rows = []

        node_names: list[str] = split_commas(
            "name,id,ha_servers,check_type,checkinter,log-health-checks,httpcheck_method,"
            "balance,balance_urilen,balance_uridepth,balance_uriwhole,"
            "a_acl,a_actionitems,errorfiles,advanced,advanced_backend,"
            "transparent_clientip,transparent_interface,"
            "monitor_uri,monitor_httpversion,monitor_username,monitor_domain,"
            "monitor_agentport,agent_check,agent_port,agent_port,"
            "connection_timeout,server_timeout,retries,"
            "stats_enabled,stats_username,stats_password,stats_uri,stats_scope,stats_realm,"
            "stats_admin,stats_node,stats_desc,stats_refresh,"
            "persist_stick_expire,persist_stick_tablesize,persist_stick_length,"
            "persist_stick_cookiename,persist_sticky_type,persist_cookie_enabled,"
            "persist_cookie_name,persist_cookie_mode,persist_cookie_cachable,"
            "persist_cookie_postonly,persist_cookie_httponly,persist_cookie_secure,"
            "haproxy_cookie_maxidle,haproxy_cookie_maxlife,haproxy_cookie_domains,"
            "haproxy_cookie_dynamic_cookie_key,strict_transport_security,"
            "cookie_attribute_secure,email_level,email_to,agent_inter"
        )

        column_widths: list[int] = split_commas(
            "20,20,80,20,20,30,25,20,20,25,"
            "25,20,20,20,20,25,25,30,20,25,"
            "25,25,20,25,20,20,20,25,20,20,"
            "25,30,25,30,25,40,25,25,30,25,"
            "30,30,30,30,30,30,30,30,30,30,"
            "30,30,30,30,30,30,30,20,20,30"
        )

        rows = []

        for node in nodes:
            self.report_unknown_node_elements(node, node_names)
            row = []

            for node_name in node_names:
                value = self.adjust_node(xml_findone(node, node_name))
                row.append(value)
            self.sanity_check_node_row(node, row)
            rows.append(row)

        yield SheetData(
            sheet_name="HAProxy Pools",
            header_row=node_names,
            data_rows=rows,
            column_widths=column_widths,
        )



"""Plugin support tools."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

import importlib
import pkgutil
from types import ModuleType
from typing import Iterator

from . import plugins
from .plugins.base_plugin import BasePlugin


def iter_namespace(ns_pkg: ModuleType) -> Iterator[pkgutil.ModuleInfo]:
    """Gather all modules in namespace."""
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


def discover_plugins() -> dict[str, BasePlugin]:
    """Discover and initialize plugins."""
    discovered_plugins: dict[str, BasePlugin] = {}
    for _, long_name, _ in iter_namespace(plugins):
        if (name := long_name.split(".")[-1]).startswith("plugin_"):
            name = name.replace("plugin_", "", 1)
            discovered_plugins[name] = importlib.import_module(long_name).Plugin()

    return discovered_plugins



"""System Users plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_node_exists

NODE_NAMES = (
    "disabled,name,groupname,scope,expires,"
    "descr,ipsecpk,uid,cert,bcrypt-hash,"
    "authorizedkeys,ipsecpsk,priv,dashboardcolumns,webguicss"
)
WIDTHS = "12,40,40,20,20,60,20,10,20,20,20,20,40,40,60"


class Plugin(BasePlugin):
    """Gather data for the System Users sheet."""

    def __init__(
        self,
        display_name: str = "System Users",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """
        Sheet with system.user information.

        Not all nodes displayed as # column on dashboard and webguicss are uninteresting
        (at least to me at the moment).
        """
        rows = []

        system_user_nodes = xml_findall(parsed_xml, "system,user")
        if not system_user_nodes:
            return

        system_user_nodes.sort(key=lambda x: x.text.casefold())

        for node in system_user_nodes:
            self.report_unknown_node_elements(node)
            row = []
            for node_name in self.node_names:
                values = [self.adjust_node(x) for x in xml_findall(node, node_name)]
                values.sort()

                row.append("\n".join(values))

            # The existence of the disabled element indicates user is disabled.
            row[1] = "Yes" if xml_node_exists(node, "disabled") else "Yes"
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""Gateways plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_findone

NODE_NAMES = (
    "name,interface,gateway,defaultgw4,defaultgw6,weight,ipprotocol,monitor_disable,"
    "action_disable,descr"
)
WIDTHS = "40,20,20,20,30," "10,20,20,20,80"


class Plugin(BasePlugin):
    """Gather data for the Gateways."""

    def __init__(
        self,
        display_name: str = "Gateways",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)
        self.default_gateways = {"defaultgw4": None, "defaultgw6": None}
        self.gateway_name = None

    def adjust_node(self, node: Node) -> str:
        """Custom node adjustments."""
        if node is None:
            return ""

        match node.tag:
            case "name":
                # Remember gateway name so we can use later for defaults.
                # name must come before defaultgwX in the nodename list.
                self.gateway_name = node.text

            case "action_disable" | "monitor_disable":
                # Existence indicates yes.
                return "YES"

        return super().adjust_node(node)

    def is_default_gw(self, ipvx: str) -> str:
        """
        Return YES if ipvx is for a default gateway, else "".

        Args:
            ipvx: ipv4 or ipv6

        Return:
            YES or "".

        """
        default = self.default_gateways[ipvx]
        if self.gateway_name and default == self.gateway_name:
            return "YES"
        return ""

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather data for Gateways."""
        rows = []

        gateways_node = xml_findone(parsed_xml, "gateways")
        if gateways_node is None:
            return

        # Load default IPV4 and IPV6 gateways.
        self.default_gateways["defaultgw4"] = self.adjust_node(
            xml_findone(gateways_node, "defaultgw4")
        )
        self.default_gateways["defaultgw6"] = self.adjust_node(
            xml_findone(gateways_node, "defaultgw6")
        )

        gateway_item_nodes = xml_findall(gateways_node, "gateway_item")
        for node in gateway_item_nodes:
            self.report_unknown_node_elements(node)
            row = []

            for node_name in self.node_names:
                value = self.adjust_node(xml_findone(node, node_name))

                if node_name in ("defaultgw4", "defaultgw6"):
                    value = self.is_default_gw(node_name)

                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""SSH data plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_findone

NODE_NAMES = "filename,xmldata"
WIDTHS = "80,20"


class Plugin(BasePlugin):
    """Gather SSH data information."""

    def __init__(
        self,
        display_name: str = "SSH Data",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather ssh data information."""
        rows = []

        sshdata_nodes = xml_findall(parsed_xml, "sshdata,sshkeyfile")
        if sshdata_nodes is None:
            return

        for node in sshdata_nodes:
            self.report_unknown_node_elements(node)
            row = []
            for node_name in self.node_names:
                value = self.adjust_node(xml_findone(node, node_name))
                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""Base plugin class."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

import datetime
from abc import ABC, abstractmethod
from typing import Generator, cast

import lxml  # nosec

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from .support.elements import nice_address_sort, unescape, xml_findall, xml_findone


def split_commas(data: str | list, make_int: bool = False) -> list[int | str]:
    """
    Create list from comma-delimited string (or list).

    If make_int is True, ensure final list contents are integers.

    Args:
        data:
            String or list of data to process.

        make_int:
            True if final list must contain only integers.

    Returns:
        list

    """
    if not data:
        # Don't mess with empty strings.
        return []

    if isinstance(data, str):
        data = data.split(",")

    if make_int:
        return [int(x) for x in data]
    # Yes, we know it is a list[str]...
    # Without cast mypy thinks we're returning list[Any].
    return cast(list[int | str], data)


def _destination_source(node: Node) -> str:
    """Format destination and source addresses/ports."""
    any_address: bool = False
    address: str = ""
    port: str = ""

    for child in node.getchildren():
        match child.tag:
            case "any":
                any_address = True
            case "address" | "network":
                address = unescape(child.text)
            case "port":
                port = unescape(child.text)
            case _:
                raise NodeError(f"Unknown tag {child.tag} in node {node.tag}.")

    result = []
    if any_address:
        if port:
            result.append(f"any:{port}")
        else:
            result.append("any")
    else:
        if address and port:
            result.append(f"{address}:{port}")
        else:
            # Ignore port without address as I don't think that is allowed.
            result.append(address)
    return "\n".join(result)


def _created_updated(node: Node) -> str:
    """Format created/updated user and date/time."""
    result = []
    for child in node.getchildren():
        match child.tag:
            case "username":
                result.append(child.text)
            case "time":
                date_time = datetime.datetime.fromtimestamp(int(child.text)).strftime(
                    "%Y-%m-%d %H-%M-%S"
                )
                result.append(date_time)
            case _:

                raise NodeError(f"Unknown tag {child.tag} in node {node.tag}.")

    return "\n".join(result)


class SheetData:
    """All information required to display a worksheet."""

    def __init__(
        self,
        *,
        sheet_name: str,
        header_row: list[str],
        data_rows: list[list],
        column_widths: list[int],
    ) -> None:
        """Gather all necessary information."""
        self.sheet_name = sheet_name
        self.header_row = header_row
        self.data_rows = data_rows
        self.column_widths = column_widths


class BasePlugin(ABC):
    """Base of all plugins."""

    def __init__(
        self,
        display_name: str,
        node_names: str,
        column_widths: str | list[int],
        el_paths_to_sanitize: list[str] | None = None,
    ) -> None:
        """
        Initialize base plugin.

        node_names:
            Comma-delimited list of nodes to obtain.
            Also used for the sheet's header row.

        column_widths:
            Comma-delimited list of sheet column widths.

        el_paths_to_sanitize:
            List of comma-delimited elements to santitize
        """
        self.display_name: str = display_name
        self.node_names: list[str] = cast(list[str], split_commas(node_names))
        self.column_widths: list[int] = cast(
            list[int], split_commas(column_widths, make_int=True)
        )
        self.el_paths_to_sanitize = el_paths_to_sanitize

    def sanitize(self, parsed_xml: Node | None) -> None:
        """
        Sanitize defined paths.

        Args:
            parsed_xml: Parsed XML Node.

        """
        if parsed_xml is None or self.el_paths_to_sanitize is None:
            # Nothing to do
            return
        assert parsed_xml is not None
        assert self.el_paths_to_sanitize is not None
        for el_path in self.el_paths_to_sanitize:
            els = xml_findall(parsed_xml, el_path)
            for el in els:
                if el.text is not None:
                    el.text = "SANITIZED"

    def adjust_node(self, node: Node) -> str | Node:
        """
        Adjust a node based children and tag name.

        Args:
            node: XML node to check

        Returns:
            Processed node data or original node if no processing done.

        """
        if node is None:
            return ""

        children = node.getchildren()
        if not children:
            # Terminal node. Defaults to returning text.
            match node.tag:
                case "address":
                    if node.text is None:
                        return ""
                    return nice_address_sort(node.text)

                case "data_ciphers" | "local_network" | "local_networkv6" | "remote_network" | "remote_networkv6":  # NOQA
                    if node.text is None:
                        return ""
                    values = [x.strip() for x in node.text.split(",")]
                    values.sort()
                    return "\n".join(values)

                case "descr":
                    if node.text is None:
                        return ""
                    value = unescape(node.text)
                    value = value.replace("<br />", "\n")
                    lines = [x.strip() for x in value.split("\n")]
                    return "\n".join(lines)

                # May be specific only to our environment. Details divided by ||
                case "detail":
                    if node.text is None:
                        return ""
                    value = unescape(node.text)
                    value = value.replace("||", "\n")
                    lines = [x.strip() for x in value.split("\n")]
                    return "\n".join(lines)

                case "disable" | "disabled" | "enable" | "blockpriv" | "blockbogons":  # NOQA
                    if node.text:
                        raise NodeError(
                            f"Node {node.tag} has unexpected text: {node.text}."
                        )

                    return "YES"

                case _:
                    return node.text or ""

        # Process
        match node.tag:
            case "destination" | "source":
                return _destination_source(node)

            case "created" | "updated":
                return _created_updated(node)

        # Not processed.
        return node

    def adjust_nodes(self, nodes: list[Node]) -> str:
        """
        Adjust multiple nodes into cell.

        All nodes have the same tag.
        # TODO: Always process for 'nodes' as we'll compress the information
        # for single instances.
        """
        if nodes is None:
            return ""

        num_nodes = len(nodes)
        if num_nodes == 0:
            return ""

        cell = []
        for node in nodes:
            cell.append(self.adjust_node(node))
            cell.append("")

        #  Remove the trailing space.
        if len(cell) > 1 and cell[-1] == "":
            cell = cell[:-1]
            self.sanity_check_node_row(node.getparent(), cell)

        return "\n".join(cell)

    def sanity_check_node_row(self, node: Node, row: list) -> None:
        """
        Ensure no row items are an XML node.

        Args:
            node: Node being processed so we can grab the tag.
            row: List items ready to write to spreadsheet

        Raises:
            NodeError

        """
        errors = []
        bad_items = [x for x in row if isinstance(x, lxml.etree._Element)]
        if not bad_items:
            return

        for item in bad_items:
            errors.append(f"Unprocessed {self.node_ancesters(node)}:{item.tag}")
            print("\n".join(errors))

    def extract_node_elements(self, node: Node) -> dict[str, str]:
        """Create dictionary of node children's tag:value."""
        data = {}
        for child in node.getchildren():
            data[child.tag] = self.adjust_node(child)
        return data

    def node_ancesters(self, node: Node) -> str:
        """Walk up through node parents."""
        path = []
        path.append(node.tag)
        while (node := node.getparent()) is not None:
            path.append(node.tag)
        path.reverse()
        return "/".join(path)

    def load_cell(self, node, node_names) -> str:
        """Load node elements into a single cell."""
        if node is None:
            return ""

        cell = []
        self.report_unknown_node_elements(node, node_names)
        for node_name in node_names:
            cell.append(
                f"{node_name}: {self.adjust_node(xml_findone(node, node_name))}"
            )
        return "\n".join(cell)

    def report_unknown_node_elements(
        self, node: Node, node_names: list[str] | None = None
    ) -> None:
        """
        Report if any unknown node elements are present.

        Report if any of the node's children's tags are not found in the nodenames list.

        """
        if node is None:
            return

        unknowns = []
        if node_names is None:
            node_names = self.node_names
        fn_set = set(node_names)

        for child in node.getchildren():
            if child.tag not in fn_set:
                unknowns.append(child.tag)

        if unknowns:
            path = self.node_ancesters(node)
            unknowns.sort()
            print(f"""Node {path} has unknown child node(s): {", ".join(unknowns)}""")

    def wip(self, node: Node) -> str:
        """Output a WIP warning."""
        print(f"WIP: {self.display_name}/{node.tag}.")
        return "WIP"

    @abstractmethod
    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """
        Run plugin.

        Args:
            pfsense:
                Root of the parsed XML file.

        Returns:
            List of rows to write to spreadsheet.

        """
        raise NotImplementedError



"""Installed Packages plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import unescape, xml_findall, xml_findone

NODE_NAMES = (
    "name,internal_name,version,descr,plugins,"
    "noembedded,logging,website,pkginfolink,filter_rule_function,"
    "configurationfile,include_file,text,tabs,"
    # package-specific information
    "ha_backends,ha_pools,config"
)
WIDTHS = "40,40,20,80,40,20,80,80,80,80,80,80,80,60,60,60,60"


def name_sort(node: Node) -> str:
    """Extract element name for sorting."""
    if node is None:
        return ""
    value = unescape(xml_findone(node, "name").text).casefold()
    return value


class Plugin(BasePlugin):
    """Gather data for the Installed Packages."""

    def __init__(
        self,
        display_name: str = "Installed Packages",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Local node customizations."""
        if node is None:
            return ""

        match node.tag:
            case "active":
                # Existence of tag indicates 'yes'.
                # Sanity check there is no text.
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"

            case "logging":
                node_names = "facilityname,logfilename,logsocket".split(",")
                return self.load_cell(node, node_names)

            case "plugins":
                # Get 'item'.
                children = node.getchildren()
                plugins = []
                for child in children:
                    assert child.tag == "item"
                    plugins.append(self.adjust_node(xml_findone(child, "type")))
                plugins.sort()
                return "\n".join(plugins)

            case "tabs":
                node_names = "name,active,tabgroup,url,text".split(",")
                tab_nodes = xml_findall(node, "tab")
                cells = []
                for tab_node in tab_nodes:
                    self.report_unknown_node_elements(tab_node, node_names)
                    cell = []
                    for node_name in node_names:
                        cell.append(
                            f"{node_name}: {self.adjust_node(xml_findone(tab_node, node_name))}"
                        )
                    cell.append("")
                    cells.append("\n".join(cell))
                if len(cells) > 0 and cells[-1] == "":
                    cells = cells[:-1]
                return "\n".join(cells)

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Document installed packages. Sort by name."""
        rows = []

        package_nodes = xml_findall(parsed_xml, "installedpackages,package")

        package_nodes.sort(
            key=name_sort,
            # key=lambda x: x.name.casefold(),
            reverse=False,
        )

        for node in package_nodes:
            self.report_unknown_node_elements(node)
            row = []

            for node_name in self.node_names:
                value = self.adjust_node(xml_findone(node, node_name))

                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""OpenVPN Server plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findone

# No named top-level nodes.
# Using top-level tags.
NODE_NAMES = "name,details"
WIDTHS = "60,100"


class Plugin(BasePlugin):
    """Gather information."""

    def __init__(
        self,
        display_name: str = "OpenVPN Server",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """
        Local adjustments.

        Note the missing steps which are not in our sample XML.
        """

        if node is None:
            return ""

        match node.tag:
            case "step1":
                node_names = "type".split(",")
                return self.load_cell(node, node_names)

            case "step6":
                node_names = (
                    "authcertca,certca,city,country,email,"
                    "keylength,lifetime,organization,state,uselist"
                ).split(",")
                return self.load_cell(node, node_names)

            case "step9":
                node_names = (
                    "authcertca,certca,authcertname,certname,city,country,email,"
                    "keylength,lifetime,organization,state,uselist"
                ).split(",")
                return self.load_cell(node, node_names)

            case "step10":
                node_names = (
                    "descr,crypto,concurrentcon,dhkey,digest,"
                    "dns1,dns2,dns3,dynip,engine,"
                    "gentlskey,interface,localnet,localport,nbttype,"
                    "protocol,tlsauth,topology,tunnelnet"
                ).split(",")
                return self.load_cell(node, node_names)

            case "step11":
                node_names = "ovpnrule,ovpnallow".split(",")
                return self.load_cell(node, node_names)

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather information."""
        rows = []

        node = xml_findone(parsed_xml, "ovpnserver")
        if node is None:
            return

        for child in node.getchildren():
            row = []
            row.append(child.tag)  # name
            row.append(self.adjust_node(child))

            self.sanity_check_node_row(child, row)
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""NTPD plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_findone

NODE_NAMES = "gps,orphan,statsgraph,interface,restrictions,ispool"
WIDTHS = "20,20,20,20,20,40"


class Plugin(BasePlugin):
    """Gather ca information."""

    def __init__(
        self,
        display_name: str = "NTPD",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Local node adjustments."""
        if node is None or node.text is None:
            return ""

        match node.tag:
            case "acl_network":
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"

            case "gps":
                field_names = "type".split(",")
                return self.load_cell(node, field_names)

            case "interface":
                # Order is important? Don't sort.
                result = node.text.split(",")
                result.sort()
                return "\n".join(result)

            case "ispool":
                # Order is important. Don't sort.
                return "\n".join(node.text.split(" "))

            case "restrictions":
                node_names = "acl_network,mask".split(",")
                row_nodes = xml_findall(node, "row")
                cell = []
                for row_node in row_nodes:
                    cell.append(self.load_cell(row_node, node_names))
                    cell.append("")
                if len(cell) > 0 and cell[-1] == "":
                    cell = cell[:-1]

                return "\n".join(cell)

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather ntpd information."""
        rows = []

        node = xml_findone(parsed_xml, "ntpd")
        if node is None:
            return

        self.report_unknown_node_elements(node)
        row = []

        for node_name in self.node_names:
            value = self.adjust_node(xml_findone(node, node_name))
            row.append(value)

        self.sanity_check_node_row(node, row)
        rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""Diag plugin.

Separate from DHCPD until we definitively understand the different elements.

"""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from .diag import Plugin  # NOQA



"""Virtual IP plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_findone

NODE_NAMES = (
    "type,descr,mode,subnet,subnet_bits,"
    "advbase,advskew,interface,password,uniqid,"
    "vhid"
)
WIDTHS = "20,80,20,20,20,20,20,40,20,20,20"


class Plugin(BasePlugin):
    """Gather virtual IP information."""

    def __init__(
        self,
        display_name: str = "Virtual IP",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather virtual IP information."""
        rows = []

        vip_nodes = xml_findall(parsed_xml, "virtualip,vip")
        if vip_nodes is None:
            return

        for node in vip_nodes:
            self.report_unknown_node_elements(node)
            row = []
            for node_name in self.node_names:
                value = self.adjust_node(xml_findone(node, node_name))
                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""System Groups plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall

NODE_NAMES = "name,type,address,url,aliasurl,updatefreq,descr,detail"
WIDTHS = "40,40,40,80,80,20,80,80"


class Plugin(BasePlugin):
    """Gather data for the System Groups."""

    def __init__(
        self,
        display_name: str = "Aliases",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Aliases sheet."""
        rows = []

        alias_nodes = xml_findall(parsed_xml, "aliases,alias")
        if not alias_nodes:
            return

        alias_nodes.sort(key=lambda x: x.text.casefold())

        for node in alias_nodes:
            self.report_unknown_node_elements(node)
            row = []
            for node_name in self.node_names:
                values = [self.adjust_node(x) for x in xml_findall(node, node_name)]
                values.sort()

                row.append("\n".join(values))

            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""Notifications plugin.

A two-column format to handle multiple notification types.
"""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findone

NODE_NAMES = "type,smtp,settings"
WIDTHS = "20,40,100"


class Plugin(BasePlugin):
    """Gather notifications information."""

    def __init__(
        self,
        display_name: str = "Notifications",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Local node adjustments.

        Only type we know about at the moment is smtp.
        """
        if node is None or node.text is None:
            return ""

        match node.tag:
            case "smtp":
                node_names = (
                    "disable,ipaddress,port,timeout,notifyemailaddress,username,password,"
                    "authentication_mechanism,fromaddress"
                ).split(",")
                cell = []
                for node_name in node_names:
                    cell.append(
                        f"{node_name}: {self.adjust_node(xml_findone(node, node_name))}"
                    )

                return "\n".join(cell)

            case "disable":
                # Existence indicates YES.
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )
                return "YES"

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather ntpd information."""
        rows = []

        node = xml_findone(parsed_xml, "notifications")
        if node is None:
            return

        self.report_unknown_node_elements(node)
        children = node.getchildren()
        if len(children) == 0:
            return

        row = []

        for child in children:
            row.append(child.tag)
            value = self.adjust_node(child)
            row.append(value)

        self.sanity_check_node_row(node, row)
        rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""PfSense class."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

import argparse
import os
from pathlib import Path
from typing import cast

from lxml import etree  # nosec
from openpyxl import Workbook
from openpyxl.styles import Border, Font, NamedStyle, PatternFill, Side
from openpyxl.styles.alignment import Alignment

from netgate_xml_to_xlsx.mytypes import Node

from .plugin_tools import discover_plugins
from .plugins.support.elements import sanitize_xml
from .spreadsheet import sheet_footer, sheet_header, write_ss_row


class PfSense:
    """Handle all pfSense parsing and conversion."""

    def __init__(self, args: argparse.Namespace, in_filename: str) -> None:
        """
        Initialize and load XML.

        Technically a bit too much work to do in an init (since it can fail).
        """
        self.args = args
        self.in_file = (in_path := Path(in_filename))
        self.raw_xml: str = ""
        self.parsed_xml: Node = None
        self.workbook: Workbook = Workbook()

        # ss_filename is expected to be overwritten by
        self.ss_output_path = self._get_output_path(in_path)
        self._init_styles()
        self.default_alignment = Alignment(wrap_text=True, vertical="top")

        self.plugins = discover_plugins()

        self._load()

    def _get_output_path(self, in_path: Path) -> Path:
        """Generate output path based on args and in_filename."""
        out_path = cast(Path, self.args.output_dir) / Path(f"{in_path.name}.xlsx")
        return out_path

    def _init_styles(self) -> None:
        """Iniitalized worksheet styles."""
        xlsx_header_font = Font(name="Calibri", size=16, italic=True, bold=True)
        xlsx_body_font = Font(name="Calibri", size=16)
        xlsx_footer_font = Font(name="Calibri", size=12, italic=True)

        body_border = Border(
            bottom=Side(border_style="dotted", color="00000000"),
            top=Side(border_style="dotted", color="00000000"),
            left=Side(border_style="dotted", color="00000000"),
            right=Side(border_style="dotted", color="00000000"),
        )

        alignment = Alignment(wrap_text=True, vertical="top")

        header = NamedStyle(name="header")
        header.alignment = alignment
        header.fill = PatternFill(
            "lightTrellis", fgColor="00339966"
        )  # fgColor="000000FF")  #fgColor="0000FF00")
        header.font = xlsx_header_font
        header.border = Border(
            bottom=Side(border_style="thin", color="00000000"),
            top=Side(border_style="thin", color="00000000"),
            left=Side(border_style="dotted", color="00000000"),
            right=Side(border_style="dotted", color="00000000"),
        )

        normal = NamedStyle(name="normal")
        normal.alignment = alignment
        normal.border = body_border
        normal.fill = PatternFill("solid", fgColor="FFFFFFFF")
        normal.font = xlsx_body_font

        footer = NamedStyle("footer")
        footer.alignment = Alignment(wrap_text=False, vertical="top")
        footer.border = body_border
        normal.fill = PatternFill("solid", fgColor="FFFFFFFF")
        footer.font = xlsx_footer_font

        self.workbook.add_named_style(header)
        self.workbook.add_named_style(normal)
        self.workbook.add_named_style(footer)

    def _sanity_check_root_node(self) -> None:
        """Check for unknown root elements."""
        expected_nodes = set(
            (
                "version,lastchange,revision,aliases,ca,"
                "cert,cron,dhcpd,dhcpdv6,diag,"
                "dnshaper,filter,gateways,hasync,ifgroups,"
                "installedpackages,interfaces,ipsec,nat,notifications,"
                "ntpd,openvpn,ovpnserver,ppps,proxyarp,"
                "rrd,shaper,snmpd,sshdata,staticroutes,"
                "switches,sysctl,syslog,system_groups,system_users,"
                "system,unbound,virtualip,vlans,widgets,"
                "wizardtemp,wol"
            ).split(",")
        )
        unknown = []
        for child in self.parsed_xml:
            if child.tag not in expected_nodes:
                unknown.append(child.tag)
        if len(unknown):
            print(f"""Unknown root node(s): {",".join(unknown)}""")

    def _load(self) -> None:
        """Load and parse Netgate xml firewall configuration.

        Return pfsense keys.
        """
        self.raw_xml = self.in_file.read_text(encoding="utf-8")
        self.parsed_xml = etree.XML(self.raw_xml)
        self._sanity_check_root_node()

    def _write_sheet(
        self,
        *,
        sheet_name: str,
        header_row: list[str],
        column_widths: list[int],
        rows: list[list],
    ) -> None:
        sheet = self.workbook.create_sheet(sheet_name)
        sheet_header(sheet, header_row, column_widths)

        # Define starting row num in case there are no rows to display.
        row_num = 2
        for row_num, row in enumerate(rows, start=row_num):
            write_ss_row(sheet, row, row_num)
        sheet_footer(sheet, row_num)

    def sanitize(self, plugins_to_run) -> None:
        """
        Sanitize the raw XML and save as original filename + '-sanitized'.

        The Netgate configuration file XML is well ordered and thus searchable via regex.

        Parse the XML into a tree and call the individual plugin sanitizers.
        Call the individual plugin sanitizers.

        Args:
            plugins_to_run: List of active plugin names.

        """
        # Run generic sanitize.
        self.raw_xml = sanitize_xml(self.raw_xml)

        # Parse xml for plugin sanitizing.
        self.parsed_xml = etree.XML(self.raw_xml)

        for plugin_name in plugins_to_run:
            plugin = self.plugins[plugin_name]
            plugin.sanitize(self.parsed_xml)

        # Pretty format XML.
        self.raw_xml = etree.tostring(self.parsed_xml, pretty_print=True).decode("utf8")

        # Save sanitized XML
        parts = os.path.splitext(self.in_file)
        if len(parts) == 1:
            out_path = Path(f"{parts[0]}-sanitized")
        else:
            out_path = Path(f"{parts[0]}-sanitized{parts[1]}")
        out_path.write_text(f"{self.raw_xml}", encoding="utf-8")
        print(f"Sanitized file written: {out_path}.")

        # Delete the unsanitized file.
        self.in_file.unlink()
        print(f"Deleted original file: {self.in_file}.")

    def run(self, plugin_name: str) -> None:
        """
        Run specific plugin and write sheet(s).

        Plugins yield a SheetData object.
        Continue iterating if it is None.
        """
        plugin = self.plugins[plugin_name]
        for sheet_data in plugin.run(self.parsed_xml):
            if sheet_data is None or not sheet_data.data_rows:
                continue

            self._write_sheet(
                sheet_name=sheet_data.sheet_name,
                header_row=sheet_data.header_row,
                column_widths=sheet_data.column_widths,
                rows=sheet_data.data_rows,
            )

    def save(self) -> None:
        """Delete empty first sheet and then save Workbook."""
        sheets = self.workbook.sheetnames
        del self.workbook[sheets[0]]
        out_path = self.ss_output_path
        self.workbook.save(out_path)



"""Script-specific exceptions."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.


class ScriptError(Exception):
    """Generic script error."""


class NodeError(ScriptError):
    """Something wrong with finding/processing a node."""



"""Parse commandline arguments."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

import argparse
import sys
from importlib.metadata import version
from pathlib import Path


def filter_infiles(in_files: list[str], include: bool = True) -> list[str]:
    """Return list of in_files that include or exclude 'sanitized'."""
    if include:
        return [x for x in in_files if "sanitized" in x]
    return [x for x in in_files if "sanitized" not in x]


def parse_args() -> argparse.Namespace:
    """
    Parse command line arguments.

    Process in_files and out_dir.
    """
    parser = argparse.ArgumentParser("Netgate XML to XLSX")
    default = "./output"
    parser.add_argument(
        "--output-dir",
        "-o",
        type=str,
        default=default,
        help=f"Output directory. Default: {default}",
    )
    parser.add_argument(
        "in_files", nargs="+", help="One or more Netgate .xml files to process."
    )
    parser.add_argument(
        "--sanitize",
        action="store_true",
        help="Sanitize the input xml files and save as <filename>-sanitized.",
    )

    __version__ = version("netgate_xml_to_xlsx")
    parser.add_argument(
        "--version",
        action="version",
        version=f"{__version__}",
        help="Show version number.",
    )

    args = parser.parse_args()

    # Filter files in/out.
    if args.sanitize:
        args.in_files = filter_infiles(args.in_files, include=False)
        msg = "All files already sanitized."
    else:
        args.in_files = filter_infiles(args.in_files)
        msg = (
            "No files contain 'sanitized' in the name.\n"
            "Run --sanitize before processing."
        )

    if not args.in_files:
        print(msg)
        sys.exit(-1)

    # Convert list of in_files to list of Path objects.
    # Ensure they are files, not directories.
    args.in_files = [Path(x) for x in args.in_files]
    args.in_files = [x for x in args.in_files if x.is_file()]

    # Convert output-dir to path and optionally create path.
    out_dir = Path(args.output_dir)
    try:
        out_dir.mkdir(parents=True, exist_ok=True)
        args.output_dir = out_dir
    except OSError as err:
        print(f"Error: {err}")
        sys.exit(-1)
    return args



"""Unbound plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_findone

NODE_NAMES = (
    "enable,hosts,domainoverrides,active_interface,outgoing_interface,"
    "custom_options,custom_options,hideversion,dnssecstripped,port,"
    "system_domain_local_zone_type,sslcertref,dnssec,tlsport,hideidentity,"
    "forwarding"
)
WIDTHS = "10,50,50,20,20,20,20,20,20,20,40,20,20,20,20,20"


class Plugin(BasePlugin):
    """Gather data Unbound."""

    def __init__(
        self,
        display_name: str = "Unbound",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Custom node adjustments."""
        if node is None:
            return ""

        match node.tag:
            case "domainoverrides" | "hosts":
                if node.tag == "domainoverrides":
                    node_names = "domain,ip,tls_hostname,descr".split(",")
                else:
                    # hosts
                    node_names = "host,domain,ip,aliases,descr".split(",")

                result = []
                for node_name in node_names:
                    child = xml_findone(node, node_name)
                    if child is None:
                        value = ""
                    else:
                        value = child.text or ""
                    result.append(f"{node_name}: {value}")
                return "\n".join(result)

            case "hideidentity" | "hideversion" | "dnssecstripped" | "dnssec" | "tlsport" | "forwarding":  # NOQA
                # Existence of tag indicates 'yes'.
                # Sanity check there is no text.
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Document unbound elements.  One row."""
        rows = []

        node = xml_findone(parsed_xml, "unbound")
        if node is None:
            return

        self.report_unknown_node_elements(node)
        row = []

        for node_name in self.node_names:
            value = self.adjust_nodes(xml_findall(node, node_name))

            row.append(value)

        self.sanity_check_node_row(node, row)
        rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""Syslog plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findone

NODE_NAMES = (
    "enable,logall,filterdescriptions,ipproto,nentries,"
    "remoteserver,remoteserver2,remoteserver3,reverse,sourceip,"
    "format,rotatecount,nologdefaultblock,nologbogons,nologprivatenets,"
    "logcompressiontype"
)

WIDTHS = "30,20,20,20,20,60,60,60,20,20,20,20,40,40,40,40"


class Plugin(BasePlugin):
    """Gather syslog information."""

    def __init__(
        self,
        display_name: str = "Syslog",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Local node adjustments."""
        if node is None:
            return ""

        match node.tag:
            case "logall" | "nologdefaultblock" | "nologbogons" | "nologprivatenets":
                # Existence of tag indicates 'yes'.
                # Sanity check there is no text.
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather syslog information."""
        rows = []

        node = xml_findone(parsed_xml, "syslog")
        if node is None:
            return

        self.report_unknown_node_elements(node)
        row = []

        for node_name in self.node_names:
            value = self.adjust_node(xml_findone(node, node_name))
            row.append(value)

        self.sanity_check_node_row(node, row)
        rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""IPSEC plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import unescape, xml_findall, xml_findone

TOP_NODE_NAMES = "async_crypto,logging,uniqueids,vtimaps"
TOP_WIDTHS = "30,30,30,30"

CLIENT_NODENAMES = ""
CLIENT_WIDTHS = ""

PHASE1_NODENAMES = (
    "descr,authentication_method,caref,certref,closeaction,"
    "dpd_delay,dpd_maxfail,encryption,ikeid,iketype,"
    "interface,lifetime,mobike,myid_data,myid_type,"
    "nat_traversal,peerid_data,peerid_type,pre-shared-key,private-key,"
    "protocol,remote-gateway,mode,rekey_time,reauth_time,"
    "rand_time,pkcs11certref,pkcs11pin"
)
PHASE1_WIDTHS = (
    "40,30,10,10,20,"  # 5
    "20,20,60,10,10,"  # 10
    "30,10,10,20,20,"  # 15
    "20,20,20,20,20,"  # 20
    "20,40,10,20,20,"  # 25
    "20,40,40"
)

PHASE2_NODENAMES = (
    "descr,encryption-algorithm-option,hash-algorithm-option,ikeid,lifetime,"
    "localid,mode,pfsgroup,pinghost,protocol,"
    "rekey_time,rand_time,remoteid,reqid,uniqid"
)

PHASE2_WIDTHS = "40,40,30,20,20,20,20,20,20,20,20,40,40,20,30"


class Plugin(BasePlugin):
    """Gather data Unbound."""

    def __init__(
        self,
        display_name: str = "IPSEC",
        node_names: str = TOP_NODE_NAMES,
        column_widths: str = TOP_WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def should_process(self, node: Node) -> bool:
        """
        Should we process the element?

        Element needs to have at least the top number of node name children + 1.
        Netgate will store several empty items.
        This seems like a quick and reasonable check.
        """
        assert node.tag in ("ipsec", "client")
        children = node.getchildren()
        if len(children) < len(self.node_names) + 1:
            return False
        return True

    def adjust_node(self, node: Node) -> str:
        """Local node adjustments."""
        if node is None:
            return ""

        match node.tag:
            case "encryption":
                cell = []
                node_names = (
                    "encryption-algorithm,hash-algorithm,dhgroup,prf-algorithm".split(
                        ","
                    )
                )
                item_nodes = xml_findall(node, "item")

                if item_nodes is None:
                    return ""

                for item_node in item_nodes:
                    cell.append(self.load_cell(item_node, node_names))
                    cell.append("")

                if len(cell) > 0 and cell[-1] == "":
                    cell = cell[:-1]
                return "\n".join(cell)

            case "encryption-algorithm":
                node_names = "name,keylen".split(",")
                return self.load_cell(node, node_names)

            case "encryption-algorithm-option":
                node_names = "name,keylen".split(",")
                return self.load_cell(node, node_names)

            case "localid":
                node_names = "type,address,netbits".split(",")
                return self.load_cell(node, node_names)

            case "logging":
                cell = []
                children = node.getchildren()
                for child in children:
                    cell.append(f"{child.tag}: {unescape(child.text)}")
                cell.sort()
                return "\n".join(cell)

            case "remoteid":
                node_names = "type,address,netbits".split(",")
                return self.load_cell(node, node_names)

            case "vtimaps":
                cell = []
                node_names = "reqid,index,ifnum".split(",")
                item_nodes = xml_findall(node, "item")
                for item_node in item_nodes:
                    cell.append(self.load_cell(item_node, node_names))
                    cell.append("")

                if len(cell) and cell[-1] == "":
                    cell = cell[:-1]

                return "\n".join(cell)

        return super().adjust_node(node)

    def gather_top(self, node: Node) -> list[str]:
        """Gather single row of top-level ipsec data."""
        rows = []

        if node is None or not self.should_process(node):
            return rows

        # Check for top-level elements plus the different phases and client.
        l_nodenames = self.node_names[:]
        l_nodenames.extend("phase1,phase2,client".split(","))
        self.report_unknown_node_elements(node, l_nodenames)
        row = []

        for node_name in self.node_names:
            value = self.adjust_node(xml_findone(node, node_name))

            row.append(value)

        self.sanity_check_node_row(node, row)

        if len(row):
            rows.append(row)

        return rows

    def gather_client(self, node_in: Node) -> list[str]:
        """Not implemented, but check if there's data."""
        client_node = xml_findone(node_in, "client")
        if client_node is None:
            return
        if self.should_process(client_node):
            print(f"WARNING: {self.node_ancesters(client_node)} is unimplemented.")

        return []

    def gather_phase1s(self, node_in: Node) -> list[str]:
        """IPSEC Phase 1 information."""
        rows = []

        phase1_nodes = xml_findall(node_in, "phase1")

        for node in phase1_nodes:
            self.report_unknown_node_elements(node)
            row = []
            for node_name in self.node_names:
                row.append(self.adjust_node(xml_findone(node, node_name)))
            self.sanity_check_node_row(node, row)
            rows.append(row)

        rows.sort()
        return rows

    def gather_phase2s(self, node_in: Node) -> list[str]:
        """IPSEC Phase 2 information."""
        rows = []

        phase2_nodes = xml_findall(node_in, "phase2")

        for node in phase2_nodes:
            self.report_unknown_node_elements(node)
            row = []
            for node_name in self.node_names:
                row.append(self.adjust_nodes(xml_findall(node, node_name)))
            self.sanity_check_node_row(node, row)
            rows.append(row)

        rows.sort()
        return rows

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Document unbound elements.  One row."""
        node = xml_findone(parsed_xml, "ipsec")
        if node is None:
            return

        keep_processing = False
        rows = self.gather_top(node)
        if rows is not None and len(rows) > 0:
            keep_processing = True
            yield SheetData(
                sheet_name=self.display_name,
                header_row=self.node_names,
                data_rows=rows,
                column_widths=self.column_widths,
            )

        if not keep_processing:
            return

        self.display_name = "IPSEC Client"
        self.node_names = CLIENT_NODENAMES.split(",")
        self.column_widths = CLIENT_WIDTHS.split(",")
        rows = self.gather_client(node)
        if rows is not None and len(rows) > 0:
            yield SheetData(
                sheet_name=self.display_name,
                header_row=self.node_names,
                data_rows=rows,
                column_widths=self.column_widths,
            )

        self.display_name = "IPSEC Phase 1"
        self.node_names = PHASE1_NODENAMES.split(",")
        self.column_widths = PHASE1_WIDTHS.split(",")
        rows = self.gather_phase1s(node)
        if rows is not None and len(rows) > 0:
            yield SheetData(
                sheet_name=self.display_name,
                header_row=self.node_names,
                data_rows=rows,
                column_widths=self.column_widths,
            )

        self.display_name = "IPSEC Phase 2"
        self.node_names = PHASE2_NODENAMES.split(",")
        self.column_widths = PHASE2_WIDTHS.split(",")
        rows = self.gather_phase2s(node)
        if rows is not None and len(rows) > 0:
            yield SheetData(
                sheet_name=self.display_name,
                header_row=self.node_names,
                data_rows=rows,
                column_widths=self.column_widths,
            )



"""WOL plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findone

NODE_NAMES = ""
WIDTHS = ""


class Plugin(BasePlugin):
    """Gather wol information."""

    def __init__(
        self,
        display_name: str = "wol",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """
        Gather wol information.

        Not yet implemented. Issue warning if it has data.
        """
        node = xml_findone(parsed_xml, "wol")
        if node is None:
            return

        self.report_unknown_node_elements(node)

        children = node.getchildren()
        if len(children) > 0:
            self.wip(node)

        yield None



"""Filter rules plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_findone

# Rules have both 'disabled' and 'enabled' entries.
# Looks like a difference between versions 21 and 22?
# TODO: Determine if they can be combined based on the version number.
NODE_NAMES = (
    "disabled,enabled,descr,source,destination,"
    "interface,direction,type,protocol,ipprotocol,"
    "quick,icmptype,statetype,statetimeout,floating,"
    "max,max-src-conn,max-src-nodes,max-src-states,os,"
    "allowopts,associated-rule-id,log,nopfsync,tag,"
    "tagged,tracker,id,created,updated"
)
WIDTHS = (
    "20,20,80,60,60,"
    "20,20,20,20,20,"
    "20,20,20,20,20,"
    "20,20,20,20,20,"
    "20,40,20,20,20,"
    "20,20,20,30,30"
)


class Plugin(BasePlugin):
    """Gather filter rules information."""

    def __init__(
        self,
        display_name: str = "Filter Rules",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Local adjustments."""
        if node is None:
            return ""

        match node.tag:
            case "source" | "destination":
                data = self.extract_node_elements(node)
                if "any" in data:
                    return "any"

                # If there's not a network, and there's not "any" there _should_ be
                # an address.
                address = data["network"] if "network" in data else data["address"]
                port = f""":{data["port"]}""" if "port" in data else ""
                return f"{address}{port}"

            case "allowopts" | "log" | "nopfsync" | "tagged":
                # Existence of tag indicates 'yes'.
                # Sanity check there is no text.
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather ssh data information."""
        rows = []

        rule_nodes = xml_findall(parsed_xml, "filter,rule")
        if rule_nodes is None:
            return

        for node in rule_nodes:
            self.report_unknown_node_elements(node)
            row = []
            for node_name in self.node_names:
                value = self.adjust_node(xml_findone(node, node_name))
                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""Main netgate converstion module."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

import sys

import toml

from netgate_xml_to_xlsx.errors import NodeError

from .parse_args import parse_args
from .pfsense import PfSense


def banner(pfsense: PfSense) -> None:
    """Tell people what we're doing."""
    print(f"Output: {pfsense.ss_output_path}.")


def _main() -> None:
    """Driver."""
    args = parse_args()
    in_files = args.in_files
    config = toml.load("./plugins.toml")

    for in_filename in in_files:
        pfsense = PfSense(args, in_filename)
        banner(pfsense)

        if args.sanitize:
            pfsense.sanitize(config["plugins"])
            continue

        # Run plugins in order.
        for plugin_to_run in config["plugins"]:
            print(f"    {plugin_to_run}")
            pfsense.run(plugin_to_run)
        pfsense.save()


def main() -> None:
    """Drive and catch exceptions."""
    try:
        _main()
    except NodeError as err:
        print(err)
        sys.exit(-1)


if __name__ == "__main__":
    main()



"""Stand-alone spreadsheet functions."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

import datetime

from openpyxl.utils import get_column_letter
from openpyxl.worksheet.worksheet import Worksheet


def sheet_header(sheet: Worksheet, columns: list, column_widths: list[int]) -> None:
    """Write header row then set the column widths."""
    write_ss_row(sheet, columns, 1, "header")

    for column_number, width in enumerate(column_widths, start=1):
        column_letter = get_column_letter(column_number)
        sheet.column_dimensions[column_letter].width = width


def write_ss_row(
    sheet: Worksheet, row: list, row_num: int, style_name: str = "normal"
) -> None:
    """
    Write a row into the spreadsheet.

    Args:
        row: A list of values to write into the row.

        row_increment: Number of rows to increment in spreadsheet before writing.

        style_name: Named XLSX style.

    Always increment the row before writing.

    """
    for column_number, value in enumerate(row, start=1):
        column_letter = get_column_letter(column_number)
        coordinate = f"{column_letter}{row_num}"
        sheet[coordinate] = value
        sheet[coordinate].style = style_name


def sheet_footer(sheet: Worksheet, row_number: int) -> None:
    """Write footer information on each sheet."""
    now = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M")
    run_date = f"Run date: {now}"

    write_ss_row(sheet, [run_date], row_number + 1, style_name="footer")



"""
Wizard Temp plugin.

No idea what this is for.
"""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findone

NODE_NAMES = "system,wangateway"
WIDTHS = "80,60"


class Plugin(BasePlugin):
    """Gather information."""

    def __init__(
        self,
        display_name: str = "Wizard Temp",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Local node adjustments."""
        if node is None or node.text is None:
            return ""

        match node.tag:
            case "system":
                node_names = "hostname,domain".split(",")
                self.report_unknown_node_elements(node, node_names)
                cell = []
                for node_name in node_names:
                    cell.append(
                        f"{node_name}: {self.adjust_node(xml_findone(node, node_name))}"
                    )

                return "\n".join(cell)

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather information."""
        rows = []

        node = xml_findone(parsed_xml, "wizardtemp")
        if node is None:
            return

        self.report_unknown_node_elements(node)
        row = []

        for node_name in self.node_names:
            value = self.adjust_node(xml_findone(node, node_name))
            row.append(value)

        self.sanity_check_node_row(node, row)
        rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""Extract elements from XML."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

import html
import ipaddress
import re

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node


def unescape(value: str | None) -> str:
    """Unescape XML entities."""
    if value is None:
        return ""
    assert value is not None

    return html.unescape(value)


def sanitize_xml(raw_xml: str) -> str:
    """Sanitize the xml."""
    regexes = (
        re.compile("(<authorizedkeys>).*?(</authorizedkeys>)"),
        re.compile("(<bcrypt-hash>).*?(</bcrypt-hash>)"),
        re.compile("(<clientcert_ca>).*?(</clientcert_ca>)"),
        re.compile("(<clientcert_crl>).*?(</clientcert_crl>)"),
        re.compile("(<ha_certificates>).*?(</ha_certificates>)"),
        re.compile("(<lighttpd_ls_password>).*?(</lighttpd_ls_password>)"),
        re.compile("(<password>).*?(</password>)"),
        re.compile("(<pre-shared-key>).*?(</pre-shared-key>)"),
        re.compile("(<prv>).*?(</prv>)"),
        re.compile("(<private-key>).*?(</private-key>)"),
        re.compile("(<radius_secret>).*?(</radius_secret>)"),
        re.compile("(<shared_key>).*?(</shared_key>)"),
        re.compile("(<ssloffloadcert>).*?(</ssloffloadcert>)"),
        re.compile("(<stats_password>).*?(</stats_password>)"),
        re.compile("(<tls>).*?(</tls>)"),
    )
    for regex in regexes:
        raw_xml = regex.sub(r"\1SANITIZED\2", raw_xml)
    return raw_xml


def is_digits_and_(data: str, and_val: str = ".") -> bool:
    """
    True if data is only digits and the and_val.
    """
    try:
        int(data)
    except ValueError:
        pass
    else:
        # Just digits
        return True

    # Get parts without the partition value.
    parts = [x for x in data.split(and_val) if x != and_val]

    try:
        [int(x) for x in parts]
    except ValueError:
        return False

    return True


def nice_address_sort(data: str, delimiter: str = " ") -> str:
    """
    Sort addresses that may consist of domains and IPv4/v6 addresses.

    Not all 'address' nodes are proper IPs and domains.
    some are ports.

    Deals with values such as 0.hostname.domain.

    TODO: Filter IPV4 and IPv6 separately.

    """
    try:
        addresses = [x.strip() for x in data.split(delimiter)]
        numeric = [x for x in addresses if len(x) > 1 and is_digits_and_(x)]
        numeric.sort(key=lambda x: ipaddress.ip_address(x.split("/")[0]))

        non_numeric = [x for x in addresses if len(x) > 1 and not is_digits_and_(x)]
        non_numeric.sort(key=str.casefold)

        result = []
        result.extend(non_numeric)
        result.extend(numeric)
        return "\n".join(result)
    except (TypeError, ValueError):
        # Not an actual address (ValueError) or mixture of IPV4 and IPV6 (TypeError).
        return data


def xml_findone(in_node: Node, el_path: str) -> Node | None:
    """
    Raise exception if more than one element is found.

    If no element is found, return None.
    """
    found = xml_findall(in_node, el_path)
    if not found:
        return None

    if len(found) > 1:
        raise NodeError(f"Found more than one result for: {el_path}.")

    return found[0]


def xml_node_exists(in_node: Node, el_path: str) -> bool:
    """
    True if node exists, else False.

    Netgate uses the existence of some XML entities to indicate value.
    """
    if in_node is None:
        return False

    node = xml_findone(in_node, el_path)
    if node is None:
        return False
    return True


def xml_findall(in_node: Node, el_path: str) -> list[Node]:
    """
    Find all instances of the XML path.

    Args:
        in_node:
            Parsed XML node.

        paths:
            Comma-delimited XML element hierarcy.
            If "pfsense" is first element, remove it.
            (etree parses "pfsense" into the "root")

        Returns:
            List of found Nodes.

    """
    path = el_path.split(",")
    if path[0] == "pfsense":
        if len(path) == 1:
            return []
        path = path[1:]
    selector = "/".join(path)

    try:
        nodes = in_node.findall(selector)
    except SyntaxError as err:
        raise NodeError(
            f"{in_node.getparent()}/{in_node.tag}. Selector: ({selector}). {err}."
        )

    return nodes



"""System plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import nice_address_sort, unescape, xml_findall, xml_findone

NODE_NAMES = (
    "aliasesresolveinterval,already_run_config_upgrade,authserver,bogons,crypto_hardware,"
    "dhcpbackup,disablebeep,disablechecksumoffloading,disablechecksumoffloading,disablelargereceiveoffloading,"  # NOQA
    "disablenatreflection,disablesegmentationoffloading,dns1host,dns2host,dnsserver,"
    "do_not_send_uniqueid,domain,firmware,gitsync,hn_altq_enable,"
    "hostname,ipv6dontcreatelocaldns,language,lastchange,logsbackup,"
    "maximumfrags,maximumstates,maximumtableentries,nextgid,nextuid,"
    "optimization,pkg_repo_conf_path,powerd_ac_mode,powerd_battery_mode,powerd_normal_mode,"
    "primaryconsole,reflectiontimeout,revision,rrdbackup,scrubrnid,serialspeed,"
    "ssh,sshguard_blocktime,sshguard_detection_time,sshguard_threshold,sshguard_whitelist,"
    "timeservers,timezone,use_mfs_tmp_size,use_mfs_var_size,version,"
    "webgui"
)
WIDTHS = "80,100"


class Plugin(BasePlugin):
    """Gather data for the System sheet."""

    def __init__(
        self,
        display_name: str = "System",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_nodes(self, nodes: list[Node]) -> str:
        """Custom node adjustments."""
        if nodes is None or len(nodes) == 0:
            return ""

        result = []
        for node in nodes:
            match node.tag:
                case "authserver":
                    node_names = (
                        "name,type,host,radius_protocol,radius_nasip_attribute,"
                        "radius_secret,radius_timeout,radius_auth_port,refid".split(",")
                    )
                    for node_name in node_names:
                        # Only the nodes that appear.
                        # Prep for other types of authentication.
                        value = self.adjust_node(xml_findone(node, node_name))
                        if value:
                            result.append(f"{node_name}: {value}")

                case "firmware":
                    node_names = "disablecheck".split(",")
                    result.append(self.load_cell(node, node_names))

                case "bogons":
                    result.append(self.adjust_node(xml_findone(node, "interval")))

                case "dnsserver":
                    result.append(unescape(node.text))

                case "gitsync":
                    node_names = "repositoryurl,branch".split(",")
                    result.append(self.load_cell(node, node_names))

                case "ssh":
                    result.append(self.adjust_node(xml_findone(node, "enabled")))

                case "timeservers":
                    time_servers = node.text or ""
                    result.append(nice_address_sort(time_servers))

                case "version":
                    if version := int(float(node.text)) < 21:
                        print(
                            f"Warning: File uses version {version}.x. "
                            "Script is only tested on XML format versions 21+."
                        )
                    result.append(node.text)

                case "webgui":
                    node_names = (
                        "althostnames,auth_refresh_time,authmode,"
                        "dashboardavailablewidgetspanel,dashboardcolumns,"
                        "interfacessort,loginautocomplete,logincss,loginshowhost,"
                        "max_procs,noantilockout,port,protocol,ssl-certref,"
                        "statusmonitoringsettingspanel,systemlogsfilterpanel,"
                        "systemlogsmanagelogpanel,webguicss,webguifixedmenu,"
                        "webguihostnamemenu"
                    ).split(",")

                    for node_name in node_names:
                        # Only the nodes that appear.
                        # Prep for other types of authentication.
                        value = self.adjust_node(xml_findone(node, node_name))
                        if value:
                            result.append(f"{node_name}: {value}")
        result.sort()

        if len(result):
            # Remove possible trailing space.
            if result[-1] == "":
                result = result[:-1]
            return "\n".join(result)

        return super().adjust_nodes(nodes)

    def adjust_node(self, node: Node) -> str:
        """Custom node adjustment."""
        if node is None:
            return ""

        # TODO: Multi-line case statement that black doesn't undo.
        match node.tag:
            case "loginshowhost" | "noantilockout" | "interfacessort" | "dashboardavailablewidgetspanel":  # NOQA
                # Existence of tag indicates 'yes'.
                # Sanity check there is no text.
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"

            case "systemlogsfilterpanel" | "systemlogsmanagelogpanel" | "statusmonitoringsettingspanel":  # NOQA
                # Existence of tag indicates 'yes'.
                # Sanity check there is no text.
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """
        System-level information.

        Only showing interesting information (at least to me at the moment).
        TODO: Needs custom report_unknown_node_elements.

        """
        rows = []

        top_level = {}
        # Version and change information is at top level.
        for key in "version,lastchange".split(","):
            top_level[key] = self.adjust_node(xml_findone(parsed_xml, key))

        system_node = xml_findone(parsed_xml, "system")
        if system_node is None:
            return

        for key in self.node_names:
            if key in top_level:
                rows.append([key, top_level[key]])
                continue

            row = [key, self.adjust_nodes(xml_findall(system_node, key))]
            self.sanity_check_node_row(system_node, row)
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row="name,value".split(","),
            data_rows=rows,
            column_widths=self.column_widths,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['netgate_xml_to_xlsx',
 'netgate_xml_to_xlsx.plugins',
 'netgate_xml_to_xlsx.plugins.plugin_aliases',
 'netgate_xml_to_xlsx.plugins.plugin_ca',
 'netgate_xml_to_xlsx.plugins.plugin_cert',
 'netgate_xml_to_xlsx.plugins.plugin_cron',
 'netgate_xml_to_xlsx.plugins.plugin_dhcpd',
 'netgate_xml_to_xlsx.plugins.plugin_dhcpdv6',
 'netgate_xml_to_xlsx.plugins.plugin_diag',
 'netgate_xml_to_xlsx.plugins.plugin_dnshaper',
 'netgate_xml_to_xlsx.plugins.plugin_filter',
 'netgate_xml_to_xlsx.plugins.plugin_gateways',
 'netgate_xml_to_xlsx.plugins.plugin_haproxy',
 'netgate_xml_to_xlsx.plugins.plugin_hasync',
 'netgate_xml_to_xlsx.plugins.plugin_ifgroups',
 'netgate_xml_to_xlsx.plugins.plugin_installedpackages',
 'netgate_xml_to_xlsx.plugins.plugin_interfaces',
 'netgate_xml_to_xlsx.plugins.plugin_ipsec',
 'netgate_xml_to_xlsx.plugins.plugin_nat',
 'netgate_xml_to_xlsx.plugins.plugin_notifications',
 'netgate_xml_to_xlsx.plugins.plugin_ntpd',
 'netgate_xml_to_xlsx.plugins.plugin_openvpn',
 'netgate_xml_to_xlsx.plugins.plugin_ovpnserver',
 'netgate_xml_to_xlsx.plugins.plugin_ppps',
 'netgate_xml_to_xlsx.plugins.plugin_proxyarp',
 'netgate_xml_to_xlsx.plugins.plugin_rrd',
 'netgate_xml_to_xlsx.plugins.plugin_shaper',
 'netgate_xml_to_xlsx.plugins.plugin_snmpd',
 'netgate_xml_to_xlsx.plugins.plugin_sshdata',
 'netgate_xml_to_xlsx.plugins.plugin_staticroutes',
 'netgate_xml_to_xlsx.plugins.plugin_switches',
 'netgate_xml_to_xlsx.plugins.plugin_sysctl',
 'netgate_xml_to_xlsx.plugins.plugin_syslog',
 'netgate_xml_to_xlsx.plugins.plugin_system',
 'netgate_xml_to_xlsx.plugins.plugin_system_groups',
 'netgate_xml_to_xlsx.plugins.plugin_system_users',
 'netgate_xml_to_xlsx.plugins.plugin_unbound',
 'netgate_xml_to_xlsx.plugins.plugin_virtualip',
 'netgate_xml_to_xlsx.plugins.plugin_vlans',
 'netgate_xml_to_xlsx.plugins.plugin_widgets',
 'netgate_xml_to_xlsx.plugins.plugin_wizardtemp',
 'netgate_xml_to_xlsx.plugins.plugin_wol',
 'netgate_xml_to_xlsx.plugins.support']

package_data = \
{'': ['*']}

install_requires = \
['defusedxml>=0.7.1,<0.8.0',
 'lxml>=4.8.0,<5.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'toml>=0.10.2,<0.11.0',
 'xmltodict>=0.13.0,<0.14.0']

entry_points = \
{'console_scripts': ['flakeheaven = flakeheaven:entrypoint',
                     'netgate-xml-to-xlsx = netgate_xml_to_xlsx.main:main']}

setup_kwargs = {
    'name': 'netgate-xml-to-xlsx',
    'version': '0.9.8',
    'description': 'Translate Netgate firewall rules to spreadsheet for review.',
    'long_description': "# Netgate Firewall Converter\n\nThe `netgate-xml-to-xlsx` application converts a standard Netgate firewall .xml configuration file to an .xlsx spreadsheet with multiple tabs.\n\nWe've run the current implementation on a handful of our own XML firewall rules, however there's are probably some complex XML elements that will throw warnings when you run on your own files.\n\nIf you have a piece of XML that doesn't parse please extract a minimal portion (starting at the pfsense root) and open a ticket (better yet, modify the plugin and provide a pull request).\n\n## What is implemented?\n* First pass of XML elements in a base Netgate firewall installation.\n* Installed packages:\n  * haproxy\n\n## Requirements\n* Python 3.10+.\n* Netgate firewall XML version 21.x or 22.x.\n\n\n## Installation\n\nRecommend installing this with pipx:\n```\npipx install netgate-xml-to-xlsx\n```\n\nOr into a virtual environment.\n\n```\npython -m pip install netgate-xml-to-xlsx\n```\n\nOnce installed, the `netgate-xml-to-xlsx` command is available on your path.\n\n## Configuration file\n\nThe script requires a configuration file called `plugins.toml` in the current working directory when you run the script.\nDownload the sample from the GitLab [repository](https://gitlab.com/appropriate-solutions-inc/netgate-xml-to-xlsx/-/blob/main/plugins.toml).\nThe `plugins.toml` file defines the plugins to run as well as the order in which they are run.\nThe default order to to run all standard plugins in alphabetical order followed by the installed packages in alphabetical order.\n\n## Usage\n\n### Help\n\n```\n# Display help\nnetgate-xml-to-xlsx --help\n```\n\n### Sanitize Before Use\nNetgate configuration files contains sensitive information.\nThe XML file must be sanitized before processing.\nThe original (unsanitized) file is deleted.\n\n```\n# Sanitize Netgate configuration file(s) for review.\nnetgate-xml-to-xlsx --sanitize firewall-config.xml\nnetgate-xml-to-xlsx --sanitize dir/*.xml\n```\n\n### Convert to Spreadsheet\n* By default, output is sent to the `./output` directory.\n* Use the `--output-dir` parameter to set a specific output directory.\n* The output filename is the input filename with `.xlsx` attached to the end.\n\n```\n# Convert a Netgate firewall configuration file.\nnetgate-xml-to-xlsx firewall-config.xml\n\n# Convert all files in a directory.\nnetgate-xml-to-xlsx /fwalls/*-sanitized.xml\n```\n\n## Implementation Notes\n\n### Plugins\n\nEach top-level (or installed package) element is implemented in a separate plugin.\nThe plugin name matches the XML element being processed.\n\nSome advantages to implementing plugins:\n\n* Simplifies testing.\n  Plugins parse XML and return a list of rows to be output.\n  Plugins do not do their own output.\n  Test provide source XML and check the returned rows.\n* There are numerous Netgate plugins which I'll probably never see.\n  You can add your own plugins, along with tests.\n\n\n## Nosec on lxml imports\nThe `#nosec` flag is added to the lxml imports as the lxml parsing is not a security concern in this environment.\n\n## asserts\nAsserts are used throughout to:\n1. provide mypy guidance\n1. check for unexpected data as we're working from XML samples and not a specification.\n\n\n### Cookiecutter References\n* [cookiecutter-hypermodern-python](https://github.com/cjolowicz/cookiecutter-hypermodern-python)\n* [cookiecutter-poetry](https://fpgmaas.github.io/cookiecutter-poetry/)\n\n\n",
    'author': 'Raymond GA Côté',
    'author_email': 'ray@AppropriateSolutions.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/appropriate-solutions-inc/netgate-xml-to-xlsx',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Interfaces plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findone

NODE_NAMES = (
    "enable,name,if,descr,alias-address,"
    "alias-subnet,spoofmac,enable,ipaddr,subnet,"
    "gateway,ipaddrv6,subnetv6,blockpriv,blockbogons,"
    "media,track6-interface,track6-prefix-id,dhcp6-duid,dhcp6-ia-pd-len,"
    "dhcp6cvpt,mediaopt,adv_dhcp6_prefix_selected_interface"
)
WIDTHS = "20,40,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,60"


class Plugin(BasePlugin):
    """Gather data for the Interfaces."""

    def __init__(
        self,
        display_name: str = "Interfaces",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Custom node adjustment."""
        if node is None:
            return ""

        match node.tag:
            case "media":
                # Existence of tag indicates 'yes'.
                # Sanity check there is no text.
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"
        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Document all interfaces."""
        rows = []

        # Don't sort interfaces. Want them in the order they are encountered.
        interfaces_node = xml_findone(parsed_xml, "interfaces")
        if interfaces_node is None:
            return

        children = interfaces_node.getchildren()
        if not len(children):
            return

        for node in children:
            self.report_unknown_node_elements(node)
            row = []

            for node_name in self.node_names:
                if node_name == "name":
                    row.append(node.tag)
                    continue
                value = self.adjust_node(xml_findone(node, node_name))

                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""DHCPD plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_findone

NODE_NAMES = (
    "name,enable,range,ddnsclientupdates,ddnsdomain,"
    "ddnsdomainkey,ddnsdomainkeyalgorithm,ddnsdomainkeyname,ddnsdomainprimary,ddnsdomainsecondary,"  # NOQA
    "defaultleasetime,denyunknown,dhcpleaseinlocaltime,dnsserver,domain,"
    "domainsearchlist,failover_peerip,filename,filename32,filename32arm,"
    "filename64,filename64arm,gateway,ignorebootp,ldap,"
    "mac_allow,mac_deny,maxleasetime,netmask,nextserver,"
    "ntpserver,numberoptions,rootpath,staticmap,tftp"
)

WIDTHS = (
    "20,20,40,40,20,"
    "40,40,40,40,40,"
    "40,40,40,40,40,"
    "40,40,40,40,20,"
    "20,20,20,20,20,"
    "20,20,20,20,20,"
    "40,20,40,60,40"
)


class Plugin(BasePlugin):
    """Gather dhcpd information."""

    def __init__(
        self,
        display_name: str = "DHCPD",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Local node adjustments."""
        if node is None:
            return ""

        match node.tag:
            case "mac_allow" | "mac_deny":
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"

            case "range":
                node_names = "from,to".split(",")
                return self.load_cell(node, node_names)

            case "staticmap":
                node_names = (
                    "descr,cid,ddnsdomain,ddnsdomainkey,ddnsdomainkeyalgorithm,"
                    "ddnsdomainkeyname,ddnsdomainprimary,ddnsdomainsecondary,"
                    "defaultleasetime,domain,domainsearchlist,"
                    "filename,filename32,filename32arm,filename64,filename64arm,"
                    "gateway,hostname,ipaddr,ldap,mac,"
                    "maxleasetime,nextserver,numberoptions,rootpath,tftp"
                )
                return self.load_cell(node, node_names.split(","))

        return super().adjust_node(node)

    def adjust_nodes(self, nodes: list[Node]) -> list[str]:
        """Local nodes adjustment."""

        if nodes is None or len(nodes) == 0:
            return ""

        match nodes[0].tag:
            case "staticmap":
                cell = []
                for node in nodes:
                    cell.append(self.adjust_node(node))
                    cell.append("")

                if len(cell) > 0 and cell[-1] == "":
                    cell = cell[:-1]
                return "\n".join(cell)
        return super().adjust_nodes(nodes)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Gather ntpd information."""
        rows = []
        dhcpd_node = xml_findone(parsed_xml, "dhcpd")
        if dhcpd_node is None:
            return

        self.report_unknown_node_elements(dhcpd_node, "lan,opt1,opt2,opt3".split(","))
        for node in dhcpd_node.getchildren():
            self.report_unknown_node_elements(node)
            row = []

            for node_name in self.node_names:
                if node_name == "name":
                    row.append(node.tag)
                    continue

                row.append(self.adjust_nodes(xml_findall(node, node_name)))

            self.sanity_check_node_row(node, row)
            rows.append(row)
        rows.sort()

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )



"""OpenVPN plugin."""
# Copyright © 2022 Appropriate Solutions, Inc. All rights reserved.

from typing import Generator

from netgate_xml_to_xlsx.errors import NodeError
from netgate_xml_to_xlsx.mytypes import Node

from ..base_plugin import BasePlugin, SheetData
from ..support.elements import xml_findall, xml_findone

NODE_NAMES = (
    "vpnid,disable,mode,protocol,dev_mode,interface,ipaddr,local_port,"
    "description,custom_options,shared_key,digest,engine,tunnel_network,"
    "tunnel_networkv6,remote_network,remote_networkv6,gwredir,gwredir6,"
    "local_network,local_networkv6,maxclients,compression,compression_push,passtos,"
    "client2client,dynamic_ip,topology,serverbridge_dhcp,serverbridge_interface,"
    "serverbridge_routegateway,serverbridge_dhcp_start,serverbridge_dhcp_end,"
    "username_as_common_name,exit_notify,sndrcvbuf,netbios_enable,netbios_ntype,"
    "netbios_scope,create_gw,verbosity_level,ncp_enable,ping_method,keepalive_interval,"
    "keepalive_timeout,ping_seconds,ping_push,ping_action,ping_action_seconds,"
    "ping_action_push,inactive_seconds,data_ciphers,data_ciphers_fallback,"
    "authmode,tls,tls_type,tlsauth_keydir,caref,certref,crlref,dh_length,"
    "ocspurl,ecdh_curve,dns_server1,cert_depth,strictusercn,allow_compression,"
    "dns_server2,dns_server3,dns_server4"
)
WIDTHS = (
    "20,20,20,20,20,30,20,20,30,20,"  # 10
    "40,20,20,30,30,30,30,20,20,30,"  # 20
    "20,20,20,30,20,20,20,20,40,40,"  # 30
    "40,40,40,50,20,20,20,20,20,20,"  # 40
    "20,20,20,30,30,20,20,20,30,30,"  # 50
    "20,20,20,20,20,40,40,40,20,20,"  # 60
    "20,20,20,20,20,20,40,20,20,20"
)


class Plugin(BasePlugin):
    """Gather data for the System Groups."""

    def __init__(
        self,
        display_name: str = "OpenVPN Servers",
        node_names: str = NODE_NAMES,
        column_widths: str = WIDTHS,
    ) -> None:
        """Initialize."""
        super().__init__(display_name, node_names, column_widths)

    def adjust_node(self, node: Node) -> str:
        """Local node adjustments."""
        if node is None:
            return ""

        match node.tag:
            case "stricusercn":
                # Existence of tag indicates 'yes'.
                # Sanity check there is no text.
                if node.text:
                    raise NodeError(
                        f"Node {node.tag} has unexpected text: {node.text}."
                    )

                return "YES"

        return super().adjust_node(node)

    def run(self, parsed_xml: Node) -> Generator[SheetData, None, None]:
        """Document all OpenVPN servers."""
        rows = []

        openvpn_server_nodes = xml_findall(parsed_xml, "openvpn,openvpn-server")
        if not len(openvpn_server_nodes):
            return

        for node in openvpn_server_nodes:
            self.report_unknown_node_elements(node)
            row = []

            for node_name in self.node_names:
                value = self.adjust_node(xml_findone(node, node_name))

                row.append(value)

            self.sanity_check_node_row(node, row)
            rows.append(row)

        yield SheetData(
            sheet_name=self.display_name,
            header_row=self.node_names,
            data_rows=rows,
            column_widths=self.column_widths,
        )
